# ExtendedAttribute.pl
#
# A Movable Type plugin tag that allows you to store extra attributes (such as 
# stylesheets, images, and additional descriptions) for anything that has a 
# defined identifier (such as blogs, categories, and entries).
#
# If you're running MySQL, you can do this same thing by adding these extra
# attributes to the database, and then using an MTSQL tag to extract the data.
# This tag, however, allows you to do this without having to know about SQL.
# 
# The key to making this work is putting the extended attribbutes in a 
# Template Module, or in a file.  The format of this extra data is standard "CSV"
# (Comma Separated Values), so that you can easily generate and keep track of these
# category attributes with a spreadsheet program (e.g. MS Excel or OpenOffice Calc).
#
# Author: Martin Sutherland (martin@sunpig.com)
#
# See http://www.sunpig.com/martin/projects/mt/extendedattribute/ for documentation,
# installation instructions, and version history
#
# This is version 1.0.1 (3 Dec 2002)
#
# Copyright (c) Sun Pig Enterprises 2002.
#

use strict;
use warnings;
use MT::Template::Context;
MT::Template::Context->add_tag(ExtendedAttribute => sub { &extendedAttribute; });


sub extendedAttribute {
	# Get the context object, and the arguments for this tag
	my ($ctx, $args) = @_;

	# Check whether a file or a module has been specified.
	# If neither has been specified, return a blank.  If both have been
	# specified, use the module instead of the file.
	
	# Get the parameters for the tag
	my $strModule = $args->{'module'} || undef;		# Module name undefined by default
	my $strFile = $args->{'file'} || undef;			# File name undefined by default
	my $exAttName = $args->{'name'} || undef;		# Extended attributes name undefined by default
	my $exAttIndex = $args->{'index'} || undef;		# Extended attribute index undefined by default
	my $hasHeaderRow = $args->{'headerrow'};
	unless (defined($hasHeaderRow)) { $hasHeaderRow = 1 };	# By default, the header row parameter is true (1)
	my $type = uc($args->{'type'}) || undef;		# Type is undefined by default

	# Check dependencies:

	# 1) There must be either a 'module' or a 'file'
	unless (defined($strFile) || defined($strModule)) { return $ctx->error("Must define either a 'file' or a 'module' parameter for use with <MTExtendedAttribute>"); }
	
	# 2) There must be a 'type'
	unless (defined($type)) { return $ctx->error("There was no 'type' defined in for <MTExtendedAttribute>"); }

	# 3) There must be a 'name' or an 'index' to look up a field in the CSV data.
	unless ( defined($exAttIndex) || defined($exAttName) )  { return $ctx->error("To look up a field, the <MTExtendedAttribute> tag must have either a 'name' or an 'index' parameter."); }

	# 4) If a 'name' has been defined, there must be a header row present (headerrow="1"), otherwise we can't look it up.  (Note that headerrow is 1 by default.)
	if ( defined($exAttName) ) {
		unless ($hasHeaderRow)  { return $ctx->error("To look up a field by 'name', the 'headerrow' parameter must be set to '1' in the <MTExtendedAttribute> tag."); }
	}

	my $id = undef;
	# Determine what type of ID we should be looking up
	SWITCH: {
		if ($type eq 'BLOG') {
			# We don't need to be in any particular context to find the blog id
			$id = $ctx->stash('blog_id');
			last SWITCH;
		}
		if ($type eq 'CATEGORY') {
			# We may be purely within the category context, or we may also
			# be inside an entry context.  If we're in an entry context, use the 
			# category ID of the entry.  Otherwise, use the category id of the 
			# category context.
			if (my $entry = $ctx->stash('entry')) {
				# The entry may have a category, but then again, maybe it doesn't
				if (my $cat = $entry->category) {
					$id = $cat->id;
				} else {
					$id = "NULL";
				}
			} elsif (my $cat = $ctx->stash('category')) {
				$id = $cat->id;
			} else {
				return $ctx->error("With a type of 'category', the <MTExtendedAttribute> tag was used outside of a category or entry context.");
			}
			last SWITCH;
		}
		if ($type eq 'ENTRY') {
			# We must be in an entry context to find the entry id
			if (my $e = $ctx->stash('entry')) {
				$id = $e->id;
			} else {
				return $ctx->error("With a type of 'entry', the <MTExtendedAttribute> tag was used outside of an entry context.");
			}
			last SWITCH;
		}
		if ($type eq 'AUTHOR') {
			# We must be in an entry context to find the entry id
			if (my $e = $ctx->stash('entry')) {
				my $auth = $e->author;
				$id = $auth->id;
			} else {
				return $ctx->error("With a type of 'author', the <MTExtendedAttribute> tag was used outside of an entry context.");
			}
			last SWITCH;
		}
		return $ctx->error("Unknown 'type' for the <MTExtendedAttribute> tag");
	}
	

	# Locate the header row (if present) and the row that matches our ID
	my $blogid = $ctx->stash('blog_id');
	my @CSVRows = getCSVRows($blogid, $strModule, $strFile, $hasHeaderRow, $id) or return $ctx->error("Unable to open the " . ($strModule ? 'module' : 'file') . " '" . ($strModule ? $strModule : $strFile) . "' specified in <MTExtendedAttribute>");
	my $headerRow = shift(@CSVRows);

	# If the returned array is only one element long, then no matching data row was found!
	# That's okay.  We just return nothing.
	my $dataRow = shift(@CSVRows) or return ('');
	
	# Get the index of the field we have to return.  (Look up the index in
	# the header row, unless we already have an index)
	unless (defined($exAttIndex)) {
		# Parse the header row
		my @headerFields = parseCSVLine($headerRow) or return $ctx->error("Unable to parse the header row of the data for <MTExtendedAttribute>");
		# Locate the attribute
		for (my $i=0; $i < @headerFields; $i++) {
			# Ignore case in looking up the header
			if (uc($headerFields[$i]) eq uc($exAttName)) {
			    $exAttIndex = $i;
			    last;
			}
		}
		unless (defined($exAttIndex)) { return $ctx->error("Unable to find the field '$exAttName' in the header row of the data for <MTExtendedAttribute>"); }
	}
	
	# We now have an index for locating the required attribute
	# Parse the relevant data line.  If we can't parse it, return an error.
	my @fields = parseCSVLine($dataRow) or return $ctx->error("Unable to parse the row containing the attributes for <MTExtendedAttribute>");
	
	# Check that the indexed field exists in the row.  No ragged-ended CSV data rows allowed!
	unless ($exAttIndex < @fields) { return $ctx->error("Not enough fields in the row containing the attributes for <MTExtendedAttribute>"); }

	# Finally, return the extended attribute value we've found.
	return ($fields[$exAttIndex]);

}


sub getCSVRows {
	# Get the input parameters
	my ($blogid, $strModule, $strFile, $hasHeaderRow, $id) = @_;
	my @CSVRows = ();

	if (defined($strModule)) {
		# get the lines from a template module

		# Use the MT::Template class
		require MT::Template;

		# Load up the template module
		my $tmpl = MT::Template->load({ name => $strModule, 
					blog_id => $blogid })
				or return(@CSVRows);

		# Turn the text in the template into an array of all the lines therein
		my $text = $tmpl->text;
		my @lines = split(/\n/, $text);

		# If there's a header row, grab it first
		if ($hasHeaderRow) {
			push @CSVRows, shift (@lines);
		} else {
			push @CSVRows, "";
		}

		# Match for the ID
		my $lineID;
		my $currentLine;
		while (@lines) {
			$currentLine = shift (@lines);
			$lineID = getIDFromLine($currentLine);
			if ($id eq $lineID) {
				push @CSVRows, $currentLine;
				# don't look any further
				last;
			}
		}

	} else {
		# get the lines from a file

		# Try to open the file
		open(CSVFH, $strFile) or return(@CSVRows);
		
		# Grab each line from the file, and add it to an array.
		if ($hasHeaderRow) {
			push @CSVRows, my $line = <CSVFH>;
		} else {
			push @CSVRows, "";
		}
	
		# Match for the ID
		my $lineID;
		my $currentLine;
		while (<CSVFH>) {
			$currentLine = $_;
			$lineID = getIDFromLine($currentLine);
			if ($id eq $lineID) {
				push @CSVRows, $currentLine;
				# don't look any further
				last;
			}
		}
		close(CSVFH);
	}

	return (@CSVRows);
}


# Return the first field of a CSV row.  (This should be the ID of the row)
# Remember to take into account the possibility that the first field may 
# be wrapped in double-quotes.
# Note that this ID field may not itself contain any commas or double-quote
# characters.
sub getIDFromLine {
	my ($line) = @_;
	$line =~ m/"??(.*?)"??,/;
	my $id = $1;
	$id =~ s/"//g;
	return($id);
}




# input: a single row from a CSV data file
# output: an array where each element is a single field from the row
sub parseCSVLine {
	
	my ($row) = shift(@_);
	# Remove line break characters, because they shag the parsing
	$row =~ s/\n|\r//g;

	# Step 1:  Split the row on commas
	my @initialSplit = split(/,/, $row);

	# Step 2:  Iterate through this array.
	# In each array element, cound the number of double-quote (") characters.
	# If an array element has an odd number of double-quotes, then somewhere further 
	# down the line in the array there must be another element, also with an odd number
	# of double-quotes.  All the array items between these two elements must be part
	# of the same data field.

	my @finalFields;
	my $fieldUnderConstruction;
	my $currentField;
	my $inTheMiddleOfAField = 0;
	while (@initialSplit) {
		$currentField = shift @initialSplit;
		if (hasOddDoubleQuotes($currentField)) {
			$inTheMiddleOfAField = !$inTheMiddleOfAField;
		}
		
		# Now, remove all single occurrances of double-quote characters ("),
		# and then replace all double instances of double-quote characters ("")
		# with single double-quote characters (").
		$currentField =~ s/"(?!")//g;
		$currentField =~ s/""/"/g;

		# Append the current field to the field under construction
		$fieldUnderConstruction .= $currentField;
		
		if (!$inTheMiddleOfAField) {
			push @finalFields, $fieldUnderConstruction;
			$fieldUnderConstruction = "";
		} else {
			# The field has comma characters in it:  make sure to preserve them.
			$fieldUnderConstruction .= ",";
		}
	}
	
	# If we've come to the end of the array, and we're still in the middle of a field, something has gone wrong.
	if ($inTheMiddleOfAField) {
		# Uh-oh
		return (undef);
	} else {
		# OK
		return (@finalFields);
	}
}


# Count the number of doublequote characters in a string, and return true if this count is odd
sub hasOddDoubleQuotes {
	my ($test) = @_;
	my $count = ($test =~ tr/"//);
	return($count % 2);
}
